@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-
 """
 C.2 The Structure of the Document (p170)
 
@@ -15,19 +13,16 @@ class document(Environment, SectionUtils):
     def title(self):
         return self.ownerDocument.userdata.get('title','')
 
+    @title.setter
+    def title(self, value):
+        self.ownerDocument.userdata["title"] = value
+
     def invoke(self, tex):
         res = Environment.invoke(self, tex)
 
         # Set initial counter values
-        if 'counters' in self.config:
-            counters = self.config['counters']
-            for name in list(counters.keys()):
-                if name.startswith(';'):
-                    continue
-                try:
-                    self.ownerDocument.context.counters[name].setcounter(counters[name]-1)
-                except TypeError:
-                    self.ownerDocument.context.counters[name].setcounter(int(counters[name])-1)
+        for name, value in self.config["counters"]["counters"].items():
+            self.ownerDocument.context.counters[name].setcounter(value-1)
 
         return res
  
