@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-
 """
 C.13.2 Space (p216)
 
@@ -32,7 +30,7 @@ class bigskipamount(DimenCommand):
 class medskipamount(DimenCommand):
     value = DimenCommand.new('12pt')
 
-class smallskip(DimenCommand):
+class smallskipamount(DimenCommand):
     value = DimenCommand.new('6pt')
 
 class addvspace(Command):
