@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-
 r"""
 Implementation of the hyperref package
 
@@ -104,16 +102,16 @@ class autoref(Command):
     args = 'label:idref'
 
 class pdfstringdef(Command):
-    args = 'macroname:string tex:string'
+    args = 'macroname:str tex:str'
 
-class textorpdfstring(Command):
-    args = 'tex:string pdf:string'
+class texorpdfstring(Command):
+    args = 'tex pdf:str'
 
 class pdfstringdefDisableCommands(Command):
-    args = 'tex:string'
+    args = 'tex:str'
 
 class hypercalcbp(Command):
-    args = 'size:string'
+    args = 'size:str'
 
 
 # Forms
