@@ -1,8 +1,6 @@
-#!/usr/bin/env python
+from plasTeX.Renderers.XHTML import _Renderer
 
-from plasTeX.Renderers.XHTML import Renderer
-
-class S5(Renderer):
+class S5(_Renderer):
     """ Renderer for S5 documents (intended for Beamer classes) """
 
 Renderer = S5 
