@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-
 import os, shutil, string, importlib
 from plasTeX.Filenames import Filenames
 from plasTeX.DOM import Node
@@ -229,7 +227,8 @@ class Renderable(object):
     def vectorImage(self):
         """ Generate a vector image and return the image filename """
         image = Node.renderer.vectorImager.getImage(self)
-        image.bitmap = Node.renderer.imager.getImage(self)
+        if Node.renderer.vectorBitmap:
+            image.bitmap = Node.renderer.imager.getImage(self)
         return image
 
     @property
@@ -299,8 +298,7 @@ class Renderable(object):
             if not hasattr(self, 'config'):
                 return
 
-            level = getattr(self, 'splitlevel',
-                            self.config['files']['split-level'])
+            level = getattr(self, 'splitlevel', r.level)
 
             # If our level doesn't invoke a split, don't return a filename
             if self.level > level:
@@ -354,6 +352,7 @@ class Renderer(dict):
     textDefault = str
     default = str
     outputType = str
+    vectorBitmap = True
     imageTypes = []
     vectorImageTypes = []
     fileExtension = ''
@@ -405,6 +404,11 @@ class Renderer(dict):
         """
         config = document.config
 
+        self.level = config["files"]["split-level"]
+        filenameTemplate = config["files"]["filename"].strip()
+        if ' ' not in filenameTemplate and '[' not in filenameTemplate:
+            self.level = -10
+
         # If there are no keys, print a warning.
         # This is most likely a problem.
         if not list(self.keys()):
@@ -416,7 +420,7 @@ class Renderer(dict):
         Node.renderer = self
 
         # Create a filename generator
-        self.newFilename = Filenames(config['files'].get('filename', raw=True),
+        self.newFilename = Filenames(config['files'].get('filename'),
                                      (config['files']['bad-chars'],
                                       config['files']['bad-chars-sub']),
                                      {'jobname':document.userdata.get('jobname', '')}, self.fileExtension)
@@ -441,7 +445,7 @@ class Renderer(dict):
             elif name == 'OSXCoreGraphics':
                 from plasTeX.Imagers.OSXCoreGraphics  import Imager
             else:
-                log.warning("Invalid imager '%s'")
+                log.warning("Invalid imager '%s'" % name)
                 continue
 
             self.imager = Imager(document, self.imageTypes)
@@ -450,6 +454,8 @@ class Renderer(dict):
             if self.imager.verify():
                 log.info('Using the imager "%s".' % name)
                 break
+            else:
+                self.imager = None
 
         # Still no imager? Just use the default.
         if self.imager is None:
@@ -472,8 +478,10 @@ class Renderer(dict):
                 break
             elif name == 'dvisvgm':
                 from plasTeX.Imagers.dvisvgm import Imager as VectorImager
+            elif name == 'pdf2svg':
+                from plasTeX.Imagers.pdf2svg import Imager as VectorImager
             else:
-                log.warning("Invalid imager '%s'")
+                log.warning("Invalid imager '%s'" % name)
                 continue
 
             self.vectorImager = VectorImager(document, self.vectorImageTypes)
@@ -506,7 +514,6 @@ class Renderer(dict):
         else:
             str(document)
 
-        # Finish rendering images
         self.imager.close()
         self.vectorImager.close()
 
