@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-
 """
 TeX
 
@@ -28,9 +26,8 @@ from plasTeX.Logging import getLogger, disableLogging, fileLogging
 __all__ = ['TeX']
 
 log = getLogger()
+
 status = getLogger('status')
-tokenlog = getLogger('parse.tokens')
-digestlog = getLogger('parse.digest')
 _type = type
 
 class bufferediter(object):
@@ -60,11 +57,11 @@ class TeX(object):
     """
     documentClass = TeXDocument
 
-    def __init__(self, ownerDocument=None, myfile=None):
+    def __init__(self, ownerDocument=None, file=None):
         if ownerDocument is None:
             ownerDocument = self.documentClass()
             self.toplevel = True
-        elif myfile:
+        elif file:
             self.toplevel = True
         else:
             self.toplevel = False
@@ -115,10 +112,10 @@ class TeX(object):
         self.currentInput = (0,0)
 
         self.jobname = None
-        if myfile is not None:
+        if file is not None:
             # Filename
-            if isinstance(myfile, (str, bytes)):
-                myfile = str(myfile)
+            if isinstance(file, (str, bytes)):
+                file = str(file)
                 '''
                 if config has no files structure
                 or encoding is not specified
@@ -133,14 +130,14 @@ class TeX(object):
                 if encoding in ['utf8', 'utf-8', 'utf_8']:
                     encoding = 'utf_8_sig'
 
-                fname = self.kpsewhich(myfile)
+                fname = self.kpsewhich(file)
                 self.input(open(fname, encoding=encoding))
-                self.jobname = os.path.basename(os.path.splitext(myfile)[0])
+                self.jobname = os.path.basename(os.path.splitext(file)[0])
 
             # File object
             else:
-                self.input(myfile)
-                self.jobname = os.path.basename(os.path.splitext(myfile.name)[0])
+                self.input(file)
+                self.jobname = os.path.basename(os.path.splitext(file.name)[0])
 
     def input(self, source):
         """
@@ -155,7 +152,7 @@ class TeX(object):
             return
         if self.jobname is None:
             if isinstance(source, str):
-                self.jobname = os.path.basename(os.path.splitext(source)[0])
+                self.jobname = ''
             elif isinstance(source, IOBase):
                 self.jobname = os.path.basename(os.path.splitext(source.name)[0])
 
@@ -170,16 +167,20 @@ class TeX(object):
 
         """
         if self.inputs:
-            self.inputs.pop()
+            old_input = self.inputs.pop()
+            try:
+                old_input.source.close()
+            except AttributeError:
+                pass
         if self.inputs:
             self.currentInput = self.inputs[-1]
 
-    def loadPackage(self, myfile, options=None):
+    def loadPackage(self, file, options=None):
         """
         Load a LaTeX package
 
         Required Arguments:
-        myfile -- name of the file to load
+        file -- name of the file to load
 
         Keyword Arguments:
         options -- options passed to the macro which is loading the package
@@ -189,7 +190,7 @@ class TeX(object):
         config = self.ownerDocument.config
 
         try:
-            path = self.kpsewhich(myfile)
+            path = self.kpsewhich(file)
         except OSError as msg:
             log.warning(msg)
             return False
@@ -206,7 +207,7 @@ class TeX(object):
                 flag = plasTeX.Command()
                 self.pushToken(flag)
                 self.input(f)
-                self.ownerDocument.context.packages[myfile] = options or {}
+                self.ownerDocument.context.packages[file] = options or {}
                 for tok in self:
                     if tok is flag:
                         break
@@ -215,7 +216,7 @@ class TeX(object):
             if msg:
                 msg = ' ( %s )' % str(msg)
             # Failed to load LaTeX style file
-            log.warning('Error opening package "%s"%s', myfile, msg)
+            log.warning('Error opening package "%s"%s', file, msg)
             status.info(' ) ')
             return False
 
@@ -276,31 +277,6 @@ class TeX(object):
             except IndexError:
                 break
 
-    def iterchars(self):
-        """
-        Iterate over input characters (untokenized)
-
-        Returns:
-        generator that iterates through the untokenized characters
-
-        """
-        # Create locals before going into generator loop
-        inputs = self.inputs
-        context = self.ownerDocument.context
-        endInput = self.endInput
-        ownerDocument = self.ownerDocument
-
-        while inputs:
-            # Walk through characters
-            try:
-                for char in inputs[-1][0].iterchars():
-                    yield char
-                else:
-                    endInput()
-            # This really shouldn't happen, but just in case...
-            except IndexError:
-                break
-
     def __iter__(self):
         """
         Iterate over tokens while expanding them
@@ -415,7 +391,6 @@ class TeX(object):
 
         return out
 
-
     def parse(self, output=None):
         """
         Parse stream content until it is empty
@@ -447,8 +422,9 @@ class TeX(object):
             raise
 
         if self.toplevel:
-            for callback in self.ownerDocument.postParseCallbacks:
-                callback()
+            for order, callbacks in sorted(self.ownerDocument.postParseCallbacks.items()):
+                for callback in callbacks:
+                    callback()
         return output
 
     def textTokens(self, text):
@@ -573,17 +549,23 @@ class TeX(object):
 
         cases = [[]]
         nesting = 0
+        correctly_terminated = False
+        iterator = self.itertokens()
 
-        for t in self.itertokens():
+        for t in iterator:
             name = getattr(t, 'macroName', '') or ''
-            if name.startswith('if'):
+            if name == 'newif':
+                cases[-1].append(t)
+                cases[-1].append(next(iterator))
+                continue
+            elif name.startswith('if'):
                 cases[-1].append(t)
                 nesting += 1
             elif name == 'fi':
-                if nesting > 1:
-                    cases[-1].append(t)
-                elif not nesting:
+                if not nesting:
+                    correctly_terminated = True
                     break
+                cases[-1].append(t)
                 nesting -= 1
             elif not(nesting) and name == 'else':
                 cases.append([])
@@ -594,6 +576,9 @@ class TeX(object):
             else:
                 cases[-1].append(t)
 
+        if not correctly_terminated:
+            log.warning(r'\end occurred when \if was incomplete')
+
         # else case for ifs without elses
         cases.append([])
 
@@ -966,7 +951,7 @@ class TeX(object):
 
         """
         argtypes = {}
-        for key, t in list(self.argtypes.items()):
+        for key, t in self.argtypes.items():
             if isinstance(t, tuple):
                 argtypes[key] = t[0]
             else:
@@ -977,7 +962,7 @@ class TeX(object):
             pass
 
         # Could not find specified type
-        elif dtype not in list(argtypes.keys()):
+        elif dtype not in argtypes.keys():
             log.warning('Could not find datatype "%s"' % dtype)
 
         # Casting to specified type
@@ -1115,7 +1100,7 @@ class TeX(object):
 
     def castDimen(self, tokens, **kwargs):
         """
-        Jain the tokens into a string and convert the result into a `dimen`
+        Join the tokens into a string and convert the result into a `dimen`
 
         Required Arguments:
         tokens -- list of tokens to cast
@@ -1134,7 +1119,7 @@ class TeX(object):
 
     def castMuDimen(self, tokens, **kwargs):
         """
-        Jain the tokens into a string and convert the result into a `MuDimen`
+        Join the tokens into a string and convert the result into a `MuDimen`
 
         Required Arguments:
         tokens -- list of tokens to cast
@@ -1153,7 +1138,7 @@ class TeX(object):
 
     def castGlue(self, tokens, **kwargs):
         """
-        Jain the tokens into a string and convert the result into a `Glue`
+        Join the tokens into a string and convert the result into a `Glue`
 
         Required Arguments:
         tokens -- list of tokens to cast
@@ -1172,7 +1157,7 @@ class TeX(object):
 
     def castMuGlue(self, tokens, **kwargs):
         """
-        Jain the tokens into a string and convert the result into a `MuGlue`
+        Join the tokens into a string and convert the result into a `MuGlue`
 
         Required Arguments:
         tokens -- list of tokens to cast
@@ -1341,15 +1326,24 @@ class TeX(object):
         # When, for example, ``\Input{name}`` is encountered, we should look in
         # the directory containing the file being processed. So the following
         # code adds the directory to the start of $TEXINPUTS.
-        TEXINPUTS = None
+        TEXINPUTS = os.environ.get("TEXINPUTS",'')
         try:
             srcDir = os.path.dirname(self.filename)
         except AttributeError:
             # I think this happens only for the command line file.
             pass
         else:
-            TEXINPUTS = os.environ.get("TEXINPUTS",'')
-            os.environ["TEXINPUTS"] = "%s%s%s%s" % (srcDir, os.path.pathsep, TEXINPUTS, os.path.pathsep)
+            if TEXINPUTS:
+                os.environ["TEXINPUTS"] = "%s%s%s%s" % (srcDir, os.path.pathsep, TEXINPUTS, os.path.pathsep)
+            else:
+                os.environ["TEXINPUTS"] = "%s%s" % (srcDir, os.path.pathsep)
+
+        def restore_texinputs():
+            # Undo any mods to $TEXINPUTS.
+            if TEXINPUTS:
+                os.environ["TEXINPUTS"] = TEXINPUTS
+            else:
+                os.environ.pop("TEXINPUTS", None)
 
         try:
             program = self.ownerDocument.config['general']['kpsewhich']
@@ -1361,6 +1355,7 @@ class TeX(object):
             output = subprocess.Popen([program, name], **kwargs).communicate()[0].strip()
             output = output.decode('utf-8')
             if output:
+                restore_texinputs()
                 return output
 
         except:
@@ -1371,13 +1366,11 @@ class TeX(object):
                     fullname = os.path.join(path,name)
                     break
             if fullname:
+                restore_texinputs()
                 return fullname
 
-        # Undo any mods to $TEXINPUTS.
-        if TEXINPUTS:
-            os.environ["TEXINPUTS"] = TEXINPUTS
-
-        raise OSError('Could not find any file named: %s' % name)
+        restore_texinputs()
+        raise FileNotFoundError('Could not find any file named: %s' % name)
 
 #
 # Parsing helper methods for parsing numbers, spaces, dimens, etc.
