@@ -1,8 +1,7 @@
-#!/usr/bin/env python
-
 from plasTeX.DOM import Node, Text
 from plasTeX import encoding
 from io import BytesIO, StringIO, TextIOWrapper
+from typing import Tuple, NewType, Optional, Callable, List, Generator
 
 # Default TeX categories
 DEFAULT_CATEGORIES = [
@@ -29,32 +28,33 @@ DEFAULT_CATEGORIES = [
 VERBATIM_CATEGORIES = [''] * 16
 VERBATIM_CATEGORIES[11] = encoding.stringletters()
 
+CatCode = NewType('CatCode', int)
 
 class Token(Text):
     """ Base class for all TeX tokens """
 
     # The 16 category codes defined by TeX
-    CC_ESCAPE = 0
-    CC_BGROUP = 1
-    CC_EGROUP = 2
-    CC_MATHSHIFT = 3
-    CC_ALIGNMENT = 4
-    CC_EOL = 5
-    CC_PARAMETER = 6
-    CC_SUPER = 7
-    CC_SUB = 8
-    CC_IGNORED = 9
-    CC_SPACE = 10
-    CC_LETTER = 11
-    CC_OTHER = 12
-    CC_ACTIVE = 13
-    CC_COMMENT = 14
-    CC_INVALID = 15
+    CC_ESCAPE = CatCode(0)
+    CC_BGROUP = CatCode(1)
+    CC_EGROUP = CatCode(2)
+    CC_MATHSHIFT = CatCode(3)
+    CC_ALIGNMENT = CatCode(4)
+    CC_EOL = CatCode(5)
+    CC_PARAMETER = CatCode(6)
+    CC_SUPER = CatCode(7)
+    CC_SUB = CatCode(8)
+    CC_IGNORED = CatCode(9)
+    CC_SPACE = CatCode(10)
+    CC_LETTER = CatCode(11)
+    CC_OTHER = CatCode(12)
+    CC_ACTIVE = CatCode(13)
+    CC_COMMENT = CatCode(14)
+    CC_INVALID = CatCode(15)
 
     TOKEN_SLOTS = __slots__ = Text.TEXT_SLOTS
 
-    catcode = None       # TeX category code
-    macroName = None     # Macro to invoke in place of this token
+    catcode = None       # type: Optional[CatCode] # TeX category code
+    macroName = None     # type: Optional[str] # Macro to invoke in place of this token
 
     def __repr__(self):
         return self.source
@@ -113,6 +113,11 @@ class EscapeSequence(Token):
     @property
     def macroName(self):
         return self
+
+    @property
+    def nodeName(self):
+        return self.macroName
+
     __slots__ = Token.TOKEN_SLOTS
 
 class BeginGroup(Token):
@@ -137,11 +142,6 @@ class Alignment(Token):
     macroName = 'active::&'
     __slots__ = Token.TOKEN_SLOTS
 
-class EndOfLine(Token):
-    catcode = Token.CC_EOL
-    isElementContentWhitespace = True
-    __slots__ = Token.TOKEN_SLOTS
-
 class Parameter(Token):
     catcode = Token.CC_PARAMETER
     __slots__ = Token.TOKEN_SLOTS
@@ -169,17 +169,6 @@ class Other(Token):
     catcode = Token.CC_OTHER
     __slots__ = Token.TOKEN_SLOTS
 
-class Active(Token):
-    catcode = Token.CC_ACTIVE
-    __slots__ = Token.TOKEN_SLOTS
-
-class Comment(Token):
-    catcode = Token.CC_COMMENT
-    nodeType = Node.COMMENT_NODE
-    nodeName = '#comment'
-    isElementContentWhitespace = True
-    __slots__ = Token.TOKEN_SLOTS
-
 class Tokenizer(object):
 
     # Tokenizer states
@@ -187,22 +176,16 @@ class Tokenizer(object):
     STATE_M = 2
     STATE_N = 4
 
-    # Array for getting token class for the corresponding catcode
-    tokenClasses = [None] * 16
-    tokenClasses[Token.CC_ESCAPE] = EscapeSequence
+    tokenClasses = [None] * 16 # type: List[Optional[Callable]]
     tokenClasses[Token.CC_BGROUP] = BeginGroup
     tokenClasses[Token.CC_EGROUP] = EndGroup
     tokenClasses[Token.CC_MATHSHIFT] = MathShift
     tokenClasses[Token.CC_ALIGNMENT] = Alignment
-    tokenClasses[Token.CC_EOL] = EndOfLine
     tokenClasses[Token.CC_PARAMETER] = Parameter
     tokenClasses[Token.CC_SUPER] = Superscript
     tokenClasses[Token.CC_SUB] = Subscript
-    tokenClasses[Token.CC_SPACE] = Space
     tokenClasses[Token.CC_LETTER] = Letter
     tokenClasses[Token.CC_OTHER] = Other
-    tokenClasses[Token.CC_ACTIVE] = Active
-    tokenClasses[Token.CC_COMMENT] = Comment
 
     def __init__(self, source, context):
         """
@@ -231,6 +214,7 @@ class Tokenizer(object):
             self.filename = '<tokens>'
         else:
             self.filename = source.name
+        self.source = source
         self.seek = source.seek
         self.read = source.read
         self.readline = source.readline
@@ -249,7 +233,7 @@ class Tokenizer(object):
             if not char or ord(char) == 10:
                 break
 
-    def iterchars(self):
+    def iterchars(self) -> Generator[Tuple[CatCode, str], None, None]:
         """
         Get the next character in the stream and its category code
 
@@ -264,7 +248,6 @@ class Tokenizer(object):
         """
         # Create locals before going into the generator loop
         mybuffer = self._charBuffer
-        classes = self.tokenClasses
         read = self.read
 
         whichCode = self.context.whichCode
@@ -308,9 +291,9 @@ class Tokenizer(object):
             if code in (CC_IGNORED, CC_INVALID):
                 continue
 
-            yield classes[code](token)
+            yield (code, token)
 
-    def pushChar(self, char):
+    def pushChar(self, char: str):
         """
         Push a character back into the stream to be re-read
 
@@ -319,6 +302,8 @@ class Tokenizer(object):
 
         """
         self._charBuffer.insert(0, char)
+        if char == '\n':
+            self.lineNumber -= 1
 
     def pushToken(self, token):
         """
@@ -357,6 +342,7 @@ class Tokenizer(object):
         global Space, EscapeSequence
         Space = Space
         EscapeSequence = EscapeSequence
+        tokenClasses = self.tokenClasses
         mybuffer = self._tokBuffer
         charIter = self.iterchars()
         context = self.context
@@ -383,18 +369,14 @@ class Tokenizer(object):
 
             # Get the next character
             try:
-                token = next(charIter)
+                (code, char) = next(charIter)
             except StopIteration:
                 return
 
-            if token.nodeType == ELEMENT_NODE:
-                raise ValueError('Expanded tokens should never make it here')
-
-            code = token.catcode
-
             # Short circuit letters and other since they are so common
             if code in (CC_LETTER, CC_OTHER):
                 self.state = STATE_M
+                token = tokenClasses[code](char)
 
             # Whitespace
             elif code == CC_SPACE:
@@ -416,7 +398,7 @@ class Tokenizer(object):
                 elif state == STATE_N:
                     # ord(token) != 10 is the same as saying token != '\n'
                     # but it is much faster.
-                    if ord(token) != 10:
+                    if ord(char) != 10:
                         self.lineNumber += 1
                         self.readline()
                     token = EscapeSequence('par')
@@ -431,26 +413,26 @@ class Tokenizer(object):
                 # Get name of command sequence
                 self.state = STATE_M
 
-                for token in charIter:
+                for (next_code, next_char) in charIter:
 
-                    if token.catcode == CC_LETTER:
-                        word = [token]
-                        for t in charIter:
-                            if t.catcode == CC_LETTER:
-                                word.append(t)
+                    if next_code == CC_LETTER:
+                        word = [next_char]
+                        for (next_code_, next_char_) in charIter:
+                            if next_code_ == CC_LETTER:
+                                word.append(next_char_)
                             else:
-                                pushChar(t)
+                                pushChar(next_char_)
                                 break
                         token = EscapeSequence(''.join(word))
 
-                    elif token.catcode == CC_EOL:
+                    elif next_code == CC_EOL:
                         #pushChar(token)
                         #token = EscapeSequence()
                         token = Space(' ')
                         self.state = STATE_S
 
                     else:
-                        token = EscapeSequence(token)
+                        token = EscapeSequence(next_char)
 #
 # Because we can implement macros both in LaTeX and Python, we don't
 # always want the whitespace to be eaten.  For example, implementing
@@ -458,7 +440,7 @@ class Tokenizer(object):
 # another macro class that would eat whitspace incorrectly.  So we
 # have to do this kind of thing in the parse() method of Macro.
 #
-                    if token.catcode != CC_EOL:
+                    if next_code != CC_EOL:
 # HACK: I couldn't get the parse() thing to work so I'm just not
 #       going to parse whitespace after EscapeSequences that end in
 #       non-letter characters as a half-assed solution.
@@ -471,7 +453,7 @@ class Tokenizer(object):
                 else: token = EscapeSequence()
 
                 # Check for any \let aliases
-                token = context.lets.get(token, token)
+                token = context.get_let(token)
 
                 # TODO: This action should be generalized so that the
                 #       tokens are processed recursively
@@ -488,11 +470,12 @@ class Tokenizer(object):
                 continue
 
             elif code == CC_ACTIVE:
-                token = EscapeSequence('active::%s' % token)
-                token = context.lets.get(token, token)
+                token = EscapeSequence('active::%s' % char)
+                token = context.get_let(token)
                 self.state = STATE_M
 
             else:
+                token = tokenClasses[code](char)
                 self.state = STATE_M
 
             prev = token
